import re
import tempfile
from pathlib import Path

from msys2dl.utilities import run_subprocess


class GpgKeybox:
    def __init__(self, location: str | Path):
        self.location = Path(location).absolute()

    def update_keys(self, key_file: Path | bytes) -> None:
        if isinstance(key_file, Path):
            key_file = key_file.read_bytes()
        text_content = key_file.decode("utf-8")

        # filter out revoked keys
        ok_blocks = []
        blocks = re.findall(
            r"-+BEGIN PGP PUBLIC KEY BLOCK-+.*?-+END PGP PUBLIC KEY BLOCK-+", text_content, flags=re.DOTALL
        )
        for block in blocks:
            if "revoked" not in block:
                ok_blocks.append(block)
        text_content = "\n".join(ok_blocks)

        # Regenerate keybox file
        with tempfile.TemporaryDirectory(prefix="msys2-keys-") as d:
            temp_file_path = Path(d) / "keys.gpg"
            temp_file_path.write_text(text_content, "utf-8")
            self._run_gpg(["--batch", "--yes", "-o", str(self.location), "--dearmor", str(temp_file_path)])

    def validate_signature(self, sig_file: Path, file: Path) -> None:
        self._run_gpg(["--verify", str(sig_file), str(file)])

    def _run_gpg(self, args: list[str]) -> None:
        run_subprocess(["gpg", "--no-default-keyring", "--keyring", str(self.location), *args])



from argparse import ArgumentParser, Namespace
from typing import Protocol

from msys2dl.application import Application


class CommandType(Protocol):
    command_name: str

    def __call__(self, app: Application, args: Namespace) -> "Command": ...

    def configure_parser(self, parser: ArgumentParser) -> None: ...


class Command:
    def __init__(self, app: Application, _args: Namespace):
        self._app: Application = app

    def run(self) -> None:
        pass

    @classmethod
    def configure_parser(cls, parser: ArgumentParser) -> None:
        pass



import re
import shutil
import tempfile
import textwrap
from argparse import Namespace
from pathlib import Path
from typing import ClassVar

from msys2dl.application import Application
from msys2dl.commands.command import Command
from msys2dl.commands.output_dir_mixin import OutputDirMixin
from msys2dl.commands.package_set_mixin import PackageSetMixin
from msys2dl.package import Package
from msys2dl.package_store import PackageFile
from msys2dl.utilities import AppError, run_subprocess


class CommandMakeDeb(PackageSetMixin, OutputDirMixin, Command):
    command_name: ClassVar[str] = "make-deb"
    action_title = "Making debian packages"

    def __init__(self, app: Application, args: Namespace):
        super().__init__(app, args)

    def do_package_action(self, package_file: PackageFile) -> None:
        deb_path = DebBuilder().build(package_file, self.output_dir)
        print(f"Generated {deb_path.name}")


class DebBuilder:
    _dir_rewrites: ClassVar[dict[str, str]] = {
        "mingw64": "usr/x86_64-w64-mingw32",
        "mingw32": "usr/i686-w64-mingw32",
    }

    def build(self, msys2_package_file: PackageFile, output_dir: Path) -> Path:
        package = msys2_package_file.metadata

        with tempfile.TemporaryDirectory(prefix=msys2_package_file.metadata.name, suffix="build") as tdir_str:
            tdir = Path(tdir_str)

            # Create build directory with the correct name for dpkg-deb
            deb_name = self._generate_package_name(package)
            build_dir = Path(tdir) / f"{deb_name}-${msys2_package_file.metadata.version}.name"

            # Extract package contents & perform directory renames
            msys2_package_file.extract(build_dir)

            # Perform directory renames
            for src, dst in self._dir_rewrites.items():
                src_path = build_dir / src
                dst_path = build_dir / dst
                if dst_path.exists():
                    raise AppError(f"unexpected directory /${dst} in MSYS2 package")
                if src_path.exists():
                    shutil.move(src_path, dst_path)

            # Perform directory renames in pkg-config files
            self._alter_paths_in_pkgconfig_files(build_dir)

            # Generate control file
            single_line_description = package.description.replace("\n", " ")
            version = self._convert_package_version(package.version)
            control_file_content = textwrap.dedent(
                f"""
               Package: {deb_name}
               Version: {version}
               Architecture: all
               Maintainer: unknown
               Description: {single_line_description}
               Recommends: {', '.join(self._generate_package_name(d) for d in package.dependencies if isinstance(d, Package))}

               """
            )
            (build_dir / "DEBIAN").mkdir(parents=True)
            (build_dir / "DEBIAN/control").write_text(control_file_content, encoding="utf-8")

            # Run dpkg-deb
            deb_file_name = f"{deb_name}_{version}_all.deb"
            deb_path = output_dir / deb_file_name
            run_subprocess(["dpkg-deb", "--root-owner-group", "-b", str(build_dir), str(deb_path)])
            return deb_path

    @classmethod
    def _alter_paths_in_pkgconfig_files(cls, root: Path) -> None:
        for pc_file in root.rglob("**/*.pc"):
            content = pc_file.read_text("utf-8")
            for src, dst in cls._dir_rewrites.items():
                content = content.replace("/" + src, "/" + dst)
            pc_file.write_text(content, "utf-8")

    @staticmethod
    def _generate_package_name(package: Package) -> str:
        return f"{package.short_name}-msys2-{package.environment.name}"

    @staticmethod
    def _convert_package_version(version: str) -> str:
        # Change version to be compatible with Debian package system
        version = version.replace("-", ".")
        version = re.sub(r"[^a-zA-Z0-9.+~]", "", version)
        if not version[0].isdigit():
            version = "0." + version
        return version



from abc import abstractmethod
from argparse import ArgumentParser, Namespace
from typing import ClassVar

from msys2dl.application import Application
from msys2dl.commands.command import Command
from msys2dl.package import Environment
from msys2dl.package_database import PackageNameResolver
from msys2dl.package_store import PackageFile
from msys2dl.progress import ProgressCounter


class PackageSetMixin(Command):
    action_title: ClassVar[str] = "action_title unset"

    def __init__(self, app: Application, args: Namespace) -> None:
        super().__init__(app, args)
        self.with_dependencies: bool = not args.no_deps
        self.default_env = Environment.by_name(args.env) if args.env else None
        self.include: list[str] = PackageNameResolver(self.default_env).resolve_full_names(args.include)
        self.exclude: list[str] = PackageNameResolver(self.default_env).resolve_full_names(args.exclude)
        self.environments = {
            Environment.by_package_name_or_raise(name) for name in (self.include + self.exclude)
        }
        self.package_files: list[PackageFile] = []
        self.check_for_conflicts = not args.ignore_conflicts

    def run(self) -> None:
        super().run()
        self._app.update_keys()
        self._app.download_databases(self.environments)
        package_set = self._app.resolve_package_set(
            self.include,
            self.exclude,
            check_conflicts=self.check_for_conflicts,
            with_dependencies=self.with_dependencies,
        )
        self.package_files = self._app.download_packages(package_set)
        with ProgressCounter(len(self.package_files), description=self.action_title) as progress:
            for package_file in self.package_files:
                self._app.check_interrupted()
                self.do_package_action(package_file)
                progress.increment()

    @abstractmethod
    def do_package_action(self, package_file: PackageFile) -> None: ...

    @classmethod
    def configure_parser(cls, parser: ArgumentParser) -> None:
        super().configure_parser(parser)
        parser.add_argument("--no-deps", action="store_true", default=False)
        parser.add_argument("--ignore-conflicts", action="store_true", default=False)
        parser.add_argument("--exclude", metavar="PACKAGE", type=str, nargs="+", default=[])
        parser.add_argument(dest="include", metavar="PACKAGE", nargs="+")
        parser.add_argument(
            "--env",
            default=None,
            choices=[n for e in Environment.all for n in [e.name, *e.alias]],
        )



from msys2dl.download.callback_registry import CallbackRegistry, InterruptFlagCallbackRegistry


class DownloadCallbacks:
    def __init__(self) -> None:
        self.complete_handlers: CallbackRegistry[[]] = CallbackRegistry()
        self.success_handlers: CallbackRegistry[[]] = CallbackRegistry()
        self.failure_handlers: CallbackRegistry[[Exception]] = CallbackRegistry()
        self.progress_handlers: CallbackRegistry[[int, int]] = CallbackRegistry()
        self.is_interrupted_handlers = InterruptFlagCallbackRegistry()

    def on_success(self) -> None:
        self.success_handlers.run_callbacks()
        self.complete_handlers.run_callbacks()

    def on_failure(self, exc: Exception) -> None:
        self.failure_handlers.run_callbacks(exc)
        self.complete_handlers.run_callbacks()

    def on_progress(self, current: int, total: int) -> None:
        self.progress_handlers.run_callbacks(current, total)

    def is_interrupted(self) -> bool:
        return self.is_interrupted_handlers.run_callbacks()

    def check_interrupted(self) -> None:
        if self.is_interrupted():
            raise InterruptedError()


class SigDownloadCallback(DownloadCallbacks):
    def __init__(self, base: DownloadCallbacks):
        super().__init__()
        self.is_interrupted_handlers.register(base.is_interrupted)



import io
import tarfile
from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path
from tarfile import TarFile, TarInfo, data_filter

from msys2dl.download.download_request import DownloadRequest
from msys2dl.package import Package
from msys2dl.utilities import decompress_zst, sanitize_file_path


@dataclass
class PackageFile:
    metadata: Package
    path: Path

    def extract(self, dst: Path) -> None:
        def need_to_extract(member: TarInfo, dest_path: str) -> TarInfo | None:
            if not data_filter(member, dest_path):
                return None
            if member.name.startswith("."):
                # No .MTREE and other pacman files
                return None
            return member

        self.as_tar_file().extractall(filter=need_to_extract, path=dst)

    def as_tar_file(self) -> TarFile:
        tar_bytes = decompress_zst(self.path.read_bytes())
        return tarfile.open(fileobj=io.BytesIO(tar_bytes), mode="r")


class PackageStore:
    def __init__(self, root: Path) -> None:
        self.root = root

    def get_package_file(self, package: Package) -> PackageFile:
        return PackageFile(package, self.path_for_package(package))

    def path_for_package(self, package: Package) -> Path:
        return self.root / package.environment.name / sanitize_file_path(package.filename)

    def make_download_request(self, base_url: str, package: Package) -> DownloadRequest:
        return DownloadRequest(
            name=str(package),
            url=base_url + package.download_path,
            dest=self.path_for_package(package),
            expected_size=package.compressed_size,
        )

    def make_download_requests(self, base_url: str, packages: Iterable[Package]) -> list[DownloadRequest]:
        return [self.make_download_request(base_url, package) for package in packages]



import os
import signal
from argparse import ArgumentParser, Namespace
from collections.abc import Iterable
from pathlib import Path
from threading import Event
from types import FrameType, TracebackType

import requests
from requests import RequestException

from msys2dl.download.download_callback import DownloadCallbacks
from msys2dl.download.download_request import DownloadRequest
from msys2dl.download.parallel_downloader import ParallelDownloader
from msys2dl.download.simple_downloader import SimpleDownloader
from msys2dl.gpg_keyring import GpgKeybox
from msys2dl.package import Environment, Package, PackageSet
from msys2dl.package_database import PackageDatabase
from msys2dl.package_store import PackageFile, PackageStore
from msys2dl.progress import DownloadProgress


class Application:
    def __init__(self, args: Namespace):
        home = Path(os.getenv("MSYS2DL_HOME") or Path("~/.local/share/msys2dl").expanduser())
        home.mkdir(parents=True, exist_ok=True)
        self._database = PackageDatabase(home / "db")
        self._package_store = PackageStore(home / "packages")
        self._keybox = GpgKeybox(home / "keybox.gpg")
        self._n_download_threads: int = args.download_threads
        self._base_url: str = args.base_url
        self._keys_url: str = args.keys_url
        self._interrupt_event: Event = Event()
        self._downloader = ParallelDownloader(
            downloader=SimpleDownloader(self._keybox), n_threads=args.download_threads
        )
        signal.signal(signal.SIGINT, self.handle_interrupt)
        signal.signal(signal.SIGTERM, self.handle_interrupt)

    def __enter__(self) -> "Application":
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self._downloader.close()
        if exc_val is not None:
            raise exc_val

    def handle_interrupt(self, _sig: int, _frame: FrameType | None) -> None:
        self._interrupt_event.set()

    def check_interrupted(self) -> None:
        if self._interrupt_event.is_set():
            raise InterruptedError()

    def update_keys(self) -> None:
        try:
            response = requests.get(self._keys_url, timeout=5)
            response.raise_for_status()
        except RequestException as exc:
            print(f"Warning: failed to update signature keys: {exc}")
            return
        self._keybox.update_keys(response.content)

    def download_databases(self, environments: Iterable[Environment], force: bool = False) -> None:
        reqs = self._database.make_download_requests(self._base_url, environments)
        self._download("Downloading package database", reqs, force)
        self._database.reload()

    def download_packages(self, packages: Iterable[Package], force: bool = False) -> list[PackageFile]:
        reqs = self._package_store.make_download_requests(self._base_url, packages)
        self._download("Downloading packages", reqs, force)
        return self.resolve_package_files(packages)

    def resolve_package_set(
        self,
        include: Iterable[str],
        exclude: Iterable[str],
        with_dependencies: bool,
        check_conflicts: bool = True,
    ) -> PackageSet:
        # Get excluded packages
        excluded_packages: list[Package] = []
        for excluded_name in exclude:
            package = self._database.get(excluded_name)
            if not package:
                print(f"Warning: unknown excluded package '{excluded_name}'")
                continue
            excluded_packages.append(package)
        # Create set from includes and excludes
        requested_packages = PackageSet(self._database.get_all_or_raise(include)) - excluded_packages
        # Add dependencies
        if with_dependencies:
            requested_packages.add_dependencies_recursively(exclude=excluded_packages)
        # Check for conflicts
        if check_conflicts:
            requested_packages.check_for_conflicts()
        return requested_packages

    def resolve_package_files(self, packages: Iterable[Package]) -> list[PackageFile]:
        return [self._package_store.get_package_file(package) for package in packages]

    def _download(self, description: str, reqs: list[DownloadRequest], force: bool = False) -> None:
        if not force:
            # Don't download if already downloaded
            reqs = [r for r in reqs if not r.dest.exists()]
        if not reqs:
            # Nothing to do
            return

        with DownloadProgress(len(reqs), description) as progress:

            def register_callbacks(request: DownloadRequest, callbacks: DownloadCallbacks) -> None:
                progress.register_callbacks(request, callbacks)
                callbacks.is_interrupted_handlers.register(self._interrupt_event.is_set)
                callbacks.success_handlers.register(lambda: print(f"Downloaded {request.name}"))

            self._downloader.execute_requests(reqs, register_callbacks=register_callbacks)

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument("--download-threads", type=int, default=5, help="Number of download threads")
        parser.add_argument("--base-url", type=str, default="https://mirror.msys2.org")
        parser.add_argument(
            "--keys-url",
            type=str,
            default="https://raw.githubusercontent.com/msys2/MSYS2-keyring/master/msys2.gpg",
        )



from argparse import ArgumentParser, Namespace
from pathlib import Path

from msys2dl.application import Application
from msys2dl.commands.command import Command


class OutputDirMixin(Command):
    def __init__(self, _app: Application, args: Namespace) -> None:
        super().__init__(_app, args)
        self.output_dir: Path = args.output

    def run(self) -> None:
        self.output_dir.mkdir(parents=True, exist_ok=True)

    @classmethod
    def configure_parser(cls, parser: ArgumentParser) -> None:
        super().configure_parser(parser)
        parser.add_argument("--output", "-o", metavar="OUTPUT_DIRECTORY", type=Path, default=Path())



import logging
from collections.abc import Callable
from typing import Generic, ParamSpec

P = ParamSpec("P")


class CallbackRegistry(Generic[P]):
    def __init__(self) -> None:
        self._callbacks: list[Callable[P, None]] = []

    def register(self, handler: Callable[P, None]) -> None:
        self._callbacks.append(handler)

    def run_callbacks(self, *args: P.args, **kwargs: P.kwargs) -> None:
        for callback in self._callbacks:
            try:
                callback(*args, **kwargs)
            except Exception as exc:
                logging.exception("Unexpected exception in callback", exc_info=exc)


class InterruptFlagCallbackRegistry:
    def __init__(self) -> None:
        self._callbacks: list[Callable[[], bool]] = []

    def register(self, handler: Callable[[], bool]) -> None:
        self._callbacks.append(handler)

    def run_callbacks(self) -> bool:
        for callback in self._callbacks:
            try:
                if callback():
                    return True
            except Exception as exc:
                logging.exception("Unexpected exception in callback", exc_info=exc)
        return False



from argparse import Namespace
from typing import ClassVar

from msys2dl.application import Application
from msys2dl.commands.command import Command
from msys2dl.commands.output_dir_mixin import OutputDirMixin
from msys2dl.commands.package_set_mixin import PackageSetMixin
from msys2dl.package_store import PackageFile


class CommandExtract(PackageSetMixin, OutputDirMixin, Command):
    command_name: ClassVar[str] = "extract"
    action_title = "Extracting files"

    def __init__(self, app: Application, args: Namespace):
        super().__init__(app, args)

    def do_package_action(self, package_file: PackageFile) -> None:
        package_file.extract(self.output_dir)
        print(f"Extracted {package_file.metadata}")



from dataclasses import dataclass
from pathlib import Path


@dataclass
class DownloadRequest:
    name: str
    url: str
    dest: Path
    expected_size: int | None = None

    @property
    def sig_url(self) -> str:
        return self.url + ".sig"

    @property
    def sig_dest(self) -> Path:
        return self.dest.with_name(self.dest.name + ".sig")

    @property
    def partial_dest(self) -> Path:
        return self.dest.with_name(self.dest.name + ".part")



import io
import tarfile
from collections.abc import Iterable
from pathlib import Path

from msys2dl.download.download_request import DownloadRequest
from msys2dl.package import Environment, Package
from msys2dl.utilities import AppError, decompress_zst


class PackageDatabase:
    def __init__(self, root: Path) -> None:
        self._root = root
        self._packages_name_dict: dict[str, "Package"] = {}
        self._packages_provides_dict: dict[str, list["Package"]] = {}
        self.reload()

    def make_download_requests(
        self, base_url: str, environments: Iterable[Environment]
    ) -> list[DownloadRequest]:
        return [
            DownloadRequest(name=e.name, url=base_url + e.database_download_path, dest=self._database_file(e))
            for e in environments
        ]

    def get(self, full_name: str) -> Package | None:
        return self._packages_name_dict.get(full_name)

    def get_or_raise(self, full_name: str) -> Package:
        package = self.get(full_name)
        if package is None:
            raise AppError(f"Package {full_name} not found")
        return package

    def get_all_or_raise(self, full_names: Iterable[str]) -> list[Package]:
        return [self.get_or_raise(name) for name in full_names]

    def reload(self) -> None:
        self._packages_name_dict = {}
        self._packages_provides_dict = {}
        # Load packages
        all_packages = []
        for env in Environment.all:
            db_file = self._database_file(env)
            if db_file.exists():
                all_packages.extend(self._load_from_file(env, db_file))
        # Populate lookup dictionaries
        for p in all_packages:
            # Add package to name lookup
            self._packages_name_dict[p.name] = p
            # Add package to provides lookup
            p_provides = {p.name, *p.provides}
            for p_name in p_provides:
                self._packages_provides_dict.setdefault(p_name, []).append(p)
        # Populate dependencies
        for p in all_packages:
            p.resolve_package_links(self._packages_name_dict, self._packages_provides_dict)

    def _database_file(self, environment: Environment) -> Path:
        return self._root / (environment.name + ".db")

    @staticmethod
    def _load_from_file(env: Environment, path: Path) -> list[Package]:
        content = decompress_zst(path.read_bytes())
        tar = tarfile.open(fileobj=io.BytesIO(content), mode="r")
        packages = []
        for member in tar.getmembers():
            file = tar.extractfile(member)
            if file and member.name.endswith("/desc"):
                packages.append(Package.from_desc(file.read().decode("utf-8"), environment=env))
        return packages


class PackageNameResolver:
    def __init__(self, default_env: Environment | None = None):
        self.default_env = default_env

    def resolve_full_name(self, name: str) -> str:
        if Environment.by_package_name(name):
            # It is full package name
            return name
        if self.default_env:
            return self.default_env.package_name_prefix + name
        raise AppError(
            f"package '{name}' does not contain environment prefix, and no default environment set"
        )

    def resolve_full_names(self, names: Iterable[str]) -> list[str]:
        return [self.resolve_full_name(name) for name in names]



import sys
from argparse import ArgumentParser

from msys2dl.application import Application
from msys2dl.commands.command import CommandType
from msys2dl.commands.command_extract import CommandExtract
from msys2dl.commands.command_make_deb import CommandMakeDeb
from msys2dl.utilities import AppError


def _run_app(argv: list[str] | None = None) -> None:
    # Configure parser
    parser = ArgumentParser()
    Application.configure_parser(parser)
    command_types: list[CommandType] = [CommandMakeDeb, CommandExtract]
    subparsers = parser.add_subparsers(dest="command_name", required=True)
    for command_type in command_types:
        subparser = subparsers.add_parser(command_type.command_name)
        command_type.configure_parser(subparser)

    # Parse command
    args = parser.parse_args(argv)

    # RUn application
    with Application(args) as app:
        command_type = next(
            command_type for command_type in command_types if command_type.command_name == args.command_name
        )
        command = command_type(app, args)
        command.run()


def main(argv: list[str] | None = None) -> None:
    try:
        _run_app(argv)
    except InterruptedError:
        print("Interrupted")
        sys.exit(2)
    except AppError as e:
        print(e.display())
        sys.exit(1)


if __name__ == "__main__":
    main()



import re
from collections import deque
from collections.abc import Iterable, Iterator
from dataclasses import dataclass, field
from typing import ClassVar, Optional

from msys2dl.utilities import AppError


@dataclass
class Environment:
    name: str
    alias: list[str]
    path_prefix: str
    package_name_prefix: str

    all: ClassVar[list["Environment"]] = []

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Environment):
            return False
        return self.name == other.name

    @property
    def database_download_path(self) -> str:
        return f"{self.path_prefix}/{self.name}/{self.name}.db"

    def package_download_path(self, filename: str) -> str:
        return f"{self.path_prefix}/{self.name}/{filename}"

    @staticmethod
    def by_name_or_raise(name: str) -> "Environment":
        env = Environment.by_name(name)
        if env:
            return env
        raise ValueError(f"Invalid environment name: {name}")

    @staticmethod
    def by_name(name: str) -> Optional["Environment"]:
        return next(
            (env for env in Environment.all if (env.name == name) or (name in env.alias)),
            None,
        )

    @staticmethod
    def by_package_name(pname: str) -> Optional["Environment"]:
        return next(
            (env for env in Environment.all if pname.startswith(env.package_name_prefix)),
            None,
        )

    @staticmethod
    def by_package_name_or_raise(pname: str) -> "Environment":
        env = Environment.by_package_name(pname)
        if env:
            return env
        raise ValueError(f"No environment in package name: {pname}")


Environment.all = [
    Environment("clangarm64", [], "/mingw", "mingw-w64-clang-aarch64-"),
    Environment("clang32", [], "/mingw", "mingw-w64-clang-i686-"),
    Environment("clang64", [], "/mingw", "mingw-w64-clang-x86_64-"),
    Environment("mingw32", [], "/mingw", "mingw-w64-i686-"),
    Environment("mingw64", [], "/mingw", "mingw-w64-x86_64-"),
    Environment("ucrt64", [], "/mingw", "mingw-w64-ucrt-x86_64-"),
]


class PackageAlternatives:
    def __init__(self, provides: str, packages: Iterable["Package"]) -> None:
        self.provides = provides
        self.packages = tuple(packages)

    def __repr__(self) -> str:
        return f"Alternatives(provides='{self.provides}', packages={self.packages})"

    def __str__(self) -> str:
        return f"providers for {self.provides} (" + ", ".join(p.name for p in self.packages) + ")"


@dataclass
class Package:
    environment: Environment
    name: str
    description: str
    version: str
    filename: str
    compressed_size: int | None
    provides: list[str]
    dependencies_str: list[str]
    conflicts_str: list[str]
    conflicts: list["Package"] = field(default_factory=list, repr=False)
    dependencies: list["Package | PackageAlternatives"] = field(default_factory=list, repr=False)
    unknown_dependencies: list[str] = field(default_factory=list)

    def __repr__(self) -> str:
        return f"Package(name='{self.name}', version={self.version})"

    def __str__(self) -> str:
        return f"{self.name}-{self.version}"

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, rhs: object) -> bool:
        if not isinstance(rhs, Package):
            return False
        return self.name == rhs.name

    @property
    def download_path(self) -> str:
        return self.environment.package_download_path(self.filename)

    @property
    def short_name(self) -> str:
        return self.name.removeprefix(self.environment.package_name_prefix)

    def resolve_package_links(
        self, name_dict: dict[str, "Package"], provides_dict: dict[str, list["Package"]]
    ) -> None:
        self.dependencies = []
        for dep in self.dependencies_str:
            if dep in provides_dict:
                provided_by = provides_dict[dep]
                self.dependencies.append(
                    provided_by[0] if len(provided_by) == 1 else PackageAlternatives(dep, provided_by)
                )
            else:
                self.unknown_dependencies.append(dep)
        for conflict_str in self.conflicts_str:
            if conflict_str in name_dict:
                self.conflicts.append(name_dict[conflict_str])

    @classmethod
    def from_desc(cls, desc: str, environment: Environment) -> "Package":
        sections = {}
        lines = desc.strip().split("\n")
        current_section_name = None
        current_section_lines = []
        for line in lines:
            if not line:
                continue
            if line.startswith("%") and line.endswith("%"):
                if current_section_name:
                    sections[current_section_name] = "\n".join(current_section_lines)
                    current_section_lines = []
                current_section_name = line[1:-1]
            else:
                current_section_lines.append(line)
        if current_section_name:
            sections[current_section_name] = "\n".join(current_section_lines)

        return Package(
            name=sections["NAME"],
            version=sections["VERSION"],
            filename=sections["FILENAME"],
            compressed_size=int(sections["CSIZE"]) if "CSIZE" in sections else None,
            description=sections.get("DESC", ""),
            conflicts_str=cls.parse_package_list(sections.get("CONFLICTS", "")),
            dependencies_str=cls.parse_package_list(sections.get("DEPENDS", "")),
            provides=cls.parse_package_list(sections.get("PROVIDES", "")),
            environment=environment,
        )

    @classmethod
    def parse_package_list(cls, text: str) -> list[str]:
        return [cls.strip_version_constraints(d) for d in text.split("\n") if d]

    @staticmethod
    def strip_version_constraints(name: str) -> str:
        return re.split(r"[<>=]", name)[0]


@dataclass
class PackageConflict:
    first: Package
    second: Package

    def __str__(self) -> str:
        return f"{self.first.name} and {self.second.name}"


class PackageSet:
    def __init__(self, packages: Iterable[Package] | None = None) -> None:
        self._set = set(packages) if packages is not None else set()

    def add(self, package: Package) -> bool:
        if package in self._set:
            return False
        self._set.add(package)
        return True

    def add_dependencies_recursively(self, exclude: Iterable[Package] | None) -> None:
        found_alternatives: set[PackageAlternatives] = set()
        q = deque(self._set)
        alternatives_q: deque[PackageAlternatives] = deque()
        while q or alternatives_q:
            if q:
                package = q.pop()
                for dep in package.dependencies:
                    if not isinstance(dep, PackageAlternatives):
                        # Normal dependency
                        if (exclude is None or dep not in exclude) and self.add(dep):
                            q.append(dep)
                    else:
                        # Alternatives present: postpone resolving them
                        # (maybe one of alternatives will be added as a normal dependency later)
                        if dep not in found_alternatives:
                            found_alternatives.add(dep)
                            alternatives_q.append(dep)
            else:
                # Need to choose from alternatives
                alternatives = alternatives_q.pop()
                chosen = next((alt for alt in alternatives.packages if alt in self._set), None)
                if chosen:
                    # already chosen
                    print(f"Alternatives: {chosen.name} is explicitly chosen from {alternatives}")
                else:
                    # select one
                    chosen = min(alternatives.packages, key=lambda p: p.name)
                    print(f"Alternatives: selecting {chosen.name} from {alternatives}")
                self.add(chosen)
                q.append(chosen)

    def find_conflicts(self) -> list[PackageConflict]:
        conflicts = []
        for p1 in self._set:
            for p2 in self._set:
                if p1 == p2 or p1.name > p2.name or (p2 not in p1.conflicts and p1 not in p2.conflicts):
                    continue

                conflicts.append(PackageConflict(p1, p2))
        return conflicts

    def check_for_conflicts(self) -> None:
        conflicts = self.find_conflicts()
        if conflicts:
            raise AppError(f"Package conflicts found: {', '.join(str(conflict) for conflict in conflicts)}")

    def __add__(self, other: Iterable[Package]) -> "PackageSet":
        return PackageSet(self._set.union(other))

    def __sub__(self, other: Iterable[Package]) -> "PackageSet":
        return PackageSet(self._set.difference(other))

    def __iter__(self) -> Iterator[Package]:
        return iter(self._set)

    def __len__(self) -> int:
        return len(self._set)



from collections.abc import Iterable

from rich import table
from rich.console import ConsoleRenderable
from rich.progress import BarColumn, DownloadColumn, MofNCompleteColumn, Progress, TaskID, TextColumn

from msys2dl.download.download_callback import DownloadCallbacks
from msys2dl.download.download_request import DownloadRequest


class ProgressCounter(Progress):
    def __init__(self, total: int, description: str = "Downloading"):
        super().__init__()
        self._main_task_id = self.add_task(description, total=total)
        self.columns = (
            TextColumn("[progress.description]{task.description}"),
            BarColumn(bar_width=None),
            MofNCompleteColumn(table_column=table.Column(width=15)),
        )

    def __enter__(self) -> "ProgressCounter":
        # override __enter__ for type checking
        super().__enter__()
        return self

    def increment(self, n: int = 1) -> None:
        self.advance(self._main_task_id, n)


class DownloadProgress(Progress):
    def __init__(self, total: int, description: str = "Downloading"):
        super().__init__()
        self._total = total
        self._main_task_id = self.add_task(description, total=total)
        self._name_column_width = len(description)

    def __enter__(self) -> "DownloadProgress":
        # override __enter__ for type checking
        super().__enter__()
        return self

    def register_callbacks(self, request: DownloadRequest, callbacks: DownloadCallbacks) -> None:
        self._name_column_width = max(self._name_column_width, len(request.name))
        DownloadProgressTracker(self, request).register_callbacks(callbacks)

    def get_renderables(self) -> Iterable[ConsoleRenderable]:
        if not self.tasks:
            return
        main_task = next(task for task in self.tasks if task.id == self._main_task_id)
        download_tasks = [task for task in self.tasks if task.id != self._main_task_id]

        # Render download progress bars
        self.columns = (
            TextColumn(
                "[progress.description]{task.description}",
                table_column=table.Column(width=self._name_column_width),
            ),
            BarColumn(bar_width=None),
            DownloadColumn(binary_units=True, table_column=table.Column(width=15)),
        )
        yield self.make_tasks_table(download_tasks)

        # Render main progress bar
        self.columns = (
            TextColumn(
                "[progress.description]{task.description}",
                table_column=table.Column(width=self._name_column_width),
            ),
            BarColumn(bar_width=None),
            MofNCompleteColumn(table_column=table.Column(width=15)),
        )
        yield self.make_tasks_table([main_task])

    def advance_n_downloaded(self, n: int = 1) -> None:
        self.advance(self._main_task_id, n)


class DownloadProgressTracker:
    def __init__(self, progress: DownloadProgress, request: DownloadRequest):
        self._task_id: TaskID | None = None
        self._progress = progress
        self._description = request.name
        self._task_removed = False

    def register_callbacks(self, callbacks: DownloadCallbacks) -> None:
        callbacks.success_handlers.register(self.on_success)
        callbacks.complete_handlers.register(self.on_complete)
        callbacks.progress_handlers.register(self.on_progress)

    def on_progress(self, bytes_downloaded: int, bytes_total: int) -> None:
        if self._task_removed:
            return
        if self._task_id is None:
            self._task_id = self._progress.add_task(self._description, total=bytes_total)
        self._progress.update(self._task_id, total=bytes_total, completed=bytes_downloaded)

    def on_success(self) -> None:
        self._progress.advance_n_downloaded(1)

    def on_complete(self) -> None:
        if not self._task_removed:
            self._task_removed = True
            if self._task_id is not None:
                self._progress.remove_task(self._task_id)



import subprocess
from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path, PurePath
from typing import Optional
from urllib.parse import quote

import zstandard as zstd
from rich.console import ConsoleRenderable
from rich.progress import (
    BarColumn,
    DownloadColumn,
    MofNCompleteColumn,
    Progress,
    TextColumn,
    TimeElapsedColumn,
)


def sanitize_file_path(path_str: str) -> PurePath:
    path = PurePath("./" + quote(path_str, safe="/"))
    if path.is_absolute():
        raise ValueError(f"Path must be relative: {path_str}")
    if any(part == ".." for part in path.parts):
        raise ValueError(f"Path must not contain ..: {path_str}")
    return path


class AppProgress(Progress):
    def get_renderables(self) -> Iterable[ConsoleRenderable]:
        for task in self.tasks:
            progress_type = task.fields.get("progress_type")
            if progress_type == "simple" or progress_type is None:
                self.columns = (
                    TextColumn("[progress.description]{task.description}"),
                    MofNCompleteColumn(),
                    BarColumn(),
                    TimeElapsedColumn(),
                )
            if progress_type == "download":
                self.columns = (
                    TextColumn("[progress.description]{task.description}"),
                    BarColumn(bar_width=None),
                    "•",
                    DownloadColumn(),
                )
            yield self.make_tasks_table([task])


def decompress_zst(inp: bytes) -> bytes:
    decompressor = zstd.ZstdDecompressor()
    stream_reader = decompressor.stream_reader(inp)
    return stream_reader.read()


def run_subprocess(command: list[str]) -> str:
    try:
        p = subprocess.run(
            command,
            check=False,
            encoding="utf-8",
            errors="ignore",
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        if p.returncode != 0:
            raise AppError(
                f"{command[0]} exited with code {p.returncode}", additional_info=["output: " + p.stdout]
            )
    except FileNotFoundError:
        raise AppError(f"no {command[0]} found in PATH")
    else:
        return p.stdout


@dataclass
class FileBlueprint:
    name: str
    content: bytes

    def write_to_dir(self, path: Path) -> None:
        (path / self.name).write_bytes(self.content)


class AppError(Exception):
    def __init__(
        self,
        message: str,
        *,
        additional_info: list[str] | None = None,
        wrapped: Optional["AppError"] = None,
    ):
        super().__init__()
        self.wrapped = wrapped
        self.message = message
        self.additional_info = list(additional_info or [])

    def display(self) -> str:
        message = ""
        combined_info_lines = []
        ex: AppError | None = self
        while ex is not None:
            if message != "":
                message += ": "
            message += ex.message
            combined_info_lines.extend(ex.additional_info)
            ex = ex.wrapped
        return "\n".join([message, *combined_info_lines])

    @classmethod
    def wrap(cls, message: str, other: "AppError") -> "AppError":
        return cls(message=message, wrapped=other)



from pathlib import Path

from requests import RequestException, Session

from msys2dl.download.download_callback import DownloadCallbacks, SigDownloadCallback
from msys2dl.download.download_request import DownloadRequest
from msys2dl.gpg_keyring import GpgKeybox
from msys2dl.utilities import AppError


class SimpleDownloader:
    def __init__(self, keybox: GpgKeybox):
        self._keybox = keybox

    def download(self, session: Session, request: DownloadRequest, callbacks: DownloadCallbacks) -> None:
        try:
            self._do_download(session, request, callbacks)
        except Exception as exc:
            callbacks.on_failure(exc)
        else:
            callbacks.on_success()

    def _do_download(self, session: Session, request: DownloadRequest, callbacks: DownloadCallbacks) -> None:
        # Download
        callbacks.check_interrupted()
        request.dest.parent.mkdir(parents=True, exist_ok=True)
        if request.expected_size is not None:
            callbacks.on_progress(0, request.expected_size)
        self._download_single_file(session, request.url, request.partial_dest, callbacks)
        self._download_single_file(session, request.sig_url, request.sig_dest, SigDownloadCallback(callbacks))
        # Check signature
        try:
            callbacks.check_interrupted()
            self._keybox.validate_signature(request.sig_dest, request.partial_dest)
        except AppError as err:
            raise AppError.wrap(f"failed to verify signature for {request.url}", err)
        # Move file to final path
        request.partial_dest.rename(request.dest)

    @staticmethod
    def _download_single_file(session: Session, url: str, dest: Path, callbacks: DownloadCallbacks) -> None:
        tries_left = 10
        last_error = None
        while tries_left > 0:
            callbacks.check_interrupted()
            tries_left -= 1
            try:
                # Start download
                with session.get(url, timeout=(5, 5), stream=True) as response:
                    if response.status_code != 200:
                        # Try again
                        last_error = "bad status code: " + str(response.status_code)
                        continue
                    # Download file
                    callbacks.check_interrupted()
                    total_size = int(response.headers.get("content-length", 0))
                    callbacks.on_progress(0, total_size)
                    bytes_downloaded = 0
                    block_size = 4 * 1024  # 4 KiB
                    with dest.open("wb") as f:
                        for data in response.iter_content(block_size):
                            callbacks.check_interrupted()
                            bytes_downloaded += len(data)
                            callbacks.on_progress(bytes_downloaded, total_size)
                            f.write(data)
                    return
            except RequestException as e:
                last_error = str(e)
        raise AppError(f"failed to download {url}: {last_error}")



from collections.abc import Callable
from concurrent.futures import CancelledError, Future, ThreadPoolExecutor
from contextvars import ContextVar

from requests import Session
from requests.adapters import HTTPAdapter

from msys2dl.download.download_callback import DownloadCallbacks
from msys2dl.download.download_request import DownloadRequest
from msys2dl.download.simple_downloader import SimpleDownloader


class Job:
    def __init__(self) -> None:
        self._failed = False
        self._exception: Exception | None = None
        self._futures: list[Future[None]] = []

    def register_callbacks(self, callback: DownloadCallbacks) -> None:
        callback.failure_handlers.register(self.on_failure)
        callback.is_interrupted_handlers.register(self.is_failed)

    def add_future(self, future: Future[None]) -> None:
        self._futures.append(future)

    def on_failure(self, exc: Exception) -> None:
        if self._failed:
            # Already failed
            return
        self._exception = exc
        self._failed = True  # interrupt all downloads
        for future in self._futures:
            future.cancel()

    def is_failed(self) -> bool:
        return self._failed

    def join(self) -> None:
        for future in self._futures:
            try:
                if not future.cancelled():
                    future.result()
            except CancelledError:
                pass
        if self._exception:
            raise self._exception


class ParallelDownloader:
    def __init__(self, *, downloader: SimpleDownloader, n_threads: int = 5) -> None:
        self._downloader = downloader
        self._pool = ThreadPoolExecutor(n_threads, initializer=self._initialize_worker_thread)
        self._session: ContextVar[Session] = ContextVar("session")
        self._sessions: list[Session] = []

    def execute_requests(
        self,
        requests: list[DownloadRequest],
        register_callbacks: Callable[[DownloadRequest, DownloadCallbacks], None],
    ) -> None:
        job = Job()
        # Add tasks to thread pool
        for request in requests:
            callbacks = DownloadCallbacks()
            register_callbacks(request, callbacks)
            job.register_callbacks(callbacks)
            job.add_future(self._pool.submit(self._execute_request, request, callbacks))
        # Wait for completion & raise exceptions if any
        job.join()

    def close(self) -> None:
        for session in self._sessions:
            session.close()
        self._pool.shutdown(wait=True)

    def _initialize_worker_thread(self) -> None:
        self._session.set(self.create_session())

    def _execute_request(self, request: DownloadRequest, callbacks: DownloadCallbacks) -> None:
        self._downloader.download(self._session.get(), request, callbacks)

    def create_session(self) -> Session:
        session = Session()
        session.mount("http://", HTTPAdapter(max_retries=0))
        session.mount("https://", HTTPAdapter(max_retries=0))
        self._sessions.append(session)
        return session
